// These classes form the basis of the RMI support
import java.rmi.RemoteException;
import java.rmi.RMISecurityManager;
import java.rmi.server.UnicastRemoteObject;
import java.io.IOException;

// We will require a number of Jini classes to support our Jini service
import net.jini.core.discovery.LookupLocator;
import net.jini.core.lookup.ServiceID;
import net.jini.discovery.DiscoveryGroupManagement;
import net.jini.discovery.LookupDiscoveryManager;
import net.jini.lookup.JoinManager;
import net.jini.lookup.ServiceIDListener;

//
// Extend the UnicastRemoteObject to create a non-activatable Jini service.
//

public class JiniAgentService extends UnicastRemoteObject
   implements Agent
{

   // The LookupDiscoveryManager you are using to find (request) LUSs
   private LookupDiscoveryManager lookupDiscMgr;

   // The lookup locator array that contains specific LUSs
   private LookupLocator locators[];

   // The groups array that contains specific groups, no groups, or all groups
   private String groups[] = DiscoveryGroupManagement.ALL_GROUPS;

   // The manager for joining LUSs
   private JoinManager joiner = null;

   // ServiceID returned from the lookup registration process
   private ServiceID serviceID = null;

   //
   // Add an init method to your constructor to do initialization specific to Jini.
   //

   public JiniAgentService() throws IOException {

   // call the UnicastRemoteObject to export the object
       super();

   // Initialize the Jini service
       init();
   }

   // Create a discovery manager by passing parameters
   // to discover all LUSes within the multicast radius.
   // Do this by specifying ALL_GROUPS and null parameters.

   private void init() throws IOException {
   
      try {
        lookupDiscMgr = new LookupDiscoveryManager(groups, locators, null);
      } catch (IOException e) {
        System.err.println("Problem starting discovery");
        e.printStackTrace();
             throw new IOException("Problem starting discovery:" +
                  e.getLocalizedMessage()); 
       }

       /* Register this service with any configured LUSes */
       if (serviceID == null) {
       // First instance ... need service id
          joiner = new JoinManager(
          this,                  // service object
          null,                  // service attributes  none at this time
          new SrvcIDListener(),  // ServiceIDListener  - internal helper class
          lookupDiscMgr,         // DiscoveryManagement - default
          null);                 // LeaseRenewalManager - default

       } else {                  // Rejoin with (recovered) state information
          joiner = new JoinManager(
          this,                  // service object
          null,                  // service attributes  none at this time
          serviceID,             // Service ID  already have an ID
          lookupDiscMgr,         // DiscoveryManagement - default
          null);                 // LeaseRenewalManager - default
       }
   }

   // implementation of the Agent interface
   public String talk() {
        return "P2P is very cool";
   }


   // The main method does not differ significantly from prior examples
   // however, you no longer need to bind to the RMIregistry because you 
   // now will use the LUS (ServiceRegistrar) to return a reference to the 
   // remote object.


   public static void main(String[] args) {
   // set the security manager
       if (System.getSecurityManager() == null) {
           System.setSecurityManager(new RMISecurityManager());
       }
       try {
            Agent agent = new JiniAgentService();
            System.out.println("JiniAgentService bound");
            Thread.currentThread().join();
       } catch (Exception e) {
            System.err.println("JiniAgentService exception: " +  e.getMessage());
            e.printStackTrace();
       }
   }

   // Utility method for setting the service's ID obtained from
   // the lookup registration process.

   private void setServiceID(ServiceID id) {
      serviceID = id;
   }


   //
   // The SrvcIDListener inner class handles the callback of the service ID 
   // assignment from the JoinManager. 
   //

   class SrvcIDListener implements ServiceIDListener
   {
   
      public SrvcIDListener() {
          super();
      }
      /**
      * The JoinManager will invoke this method when it receives a
      * valid ServiceID from a LUS.
      */
        public void serviceIDNotify(ServiceID id) {
           // Set the ID
           setServiceID(id);
           System.out.println("Received service id");    
       }
   }
}
