package server;

import java.util.*;
import java.io.*;
import net.jxta.id.IDFactory;
import net.jxta.pipe.InputPipe;
import net.jxta.pipe.PipeMsgEvent;
import net.jxta.pipe.PipeMsgListener;
import net.jxta.credential.AuthenticationCredential;
import net.jxta.document.StructuredDocument;
import net.jxta.document.Document;
import net.jxta.document.StructuredTextDocument;
import net.jxta.document.MimeMediaType;
import net.jxta.membership.Authenticator;
import net.jxta.membership.MembershipService;
import net.jxta.peergroup.PeerGroup;
import net.jxta.peergroup.PeerGroupFactory;
import net.jxta.exception.PeerGroupException;
import net.jxta.exception.ProtocolNotSupportedException;
import net.jxta.document.AdvertisementFactory;
import net.jxta.document.MimeMediaType;
import net.jxta.discovery.DiscoveryService;
import net.jxta.discovery.DiscoveryListener;
import net.jxta.discovery.DiscoveryEvent;
import net.jxta.pipe.PipeService;
import net.jxta.pipe.OutputPipe;
import net.jxta.pipe.OutputPipeListener;
import net.jxta.pipe.OutputPipeEvent;
import net.jxta.protocol.DiscoveryResponseMsg;
import net.jxta.protocol.PipeAdvertisement;
import net.jxta.protocol.PeerGroupAdvertisement;
import net.jxta.protocol.ModuleImplAdvertisement;
import net.jxta.protocol.PeerAdvertisement;
import net.jxta.protocol.ResolverQueryMsg;
import net.jxta.protocol.ResolverResponseMsg;
import net.jxta.resolver.ResolverService;
import net.jxta.resolver.QueryHandler;
import net.jxta.endpoint.Message;
import net.jxta.endpoint.MessageElement;
import net.jxta.document.Advertisement;
import net.jxta.id.ID;
import net.jxta.impl.protocol.ResolverResponse;
import net.jxta.impl.protocol.ResolverQuery;


/**
 * Encapsulates an Agent framework that handles communication between many 
 * Student peers.  Sends a series of questions to Student peers, and
 * processes their answers.  Once completed, this Teacher peer will
 * recommend additional resources to the student for improvement
 * depending on the exam result.
 */
public class Teacher {
    public static final String DISTANCE_LEARNING = "DL:";
    public static final String TEACHER_PIPE = "teacher_pipe";
    public static final String ACTION_TAG = "tag";
    public static final String CLIENT_NAME = "client";
    public static final String COURSE_NAME = "course";
    private PeerGroup netPeerGroup;
    private DiscoveryService discoverySvc;
    private PipeAdvertisement pipeAdv;
    private InputPipe inputPipe;
    private AgentFactory agentFactory = AgentFactory.getInstance();

    // This Hashtable is used to store the agents used for individual students.
    private Hashtable agentTable = new Hashtable();

    private void sendMessage(String clientName, Map map)
                      throws IOException {
        for (;;) {
            Enumeration enum = discoverySvc.getLocalAdvertisements(
                                       DiscoveryService.ADV, "Name", 
                                       Teacher.DISTANCE_LEARNING + 
                                       clientName);

            while (enum.hasMoreElements()) {
                PipeAdvertisement pipeAdv = (PipeAdvertisement) enum.nextElement();
                PipeService pipeSvc = netPeerGroup.getPipeService();

                OutputPipe outputPipe = null;

                try {
                    outputPipe = pipeSvc.createOutputPipe(pipeAdv, 1000);

                    Message message = pipeSvc.createMessage();

                    Set keySet = map.keySet();
                    Iterator iterator = keySet.iterator();

                    while (iterator.hasNext()) {
                        String key = (String) iterator.next();
                        String value = (String) map.get(key);

                        MessageElement element = message.newMessageElement(key, 
                                                                           new MimeMediaType(
                                                                                   "text/plain"), 
                                                                           new ByteArrayInputStream(
                                                                                   value.getBytes()));
                        message.addElement(element);
                    }

                    outputPipe.send(message);
                    System.out.println("Sent message");

                    return;
                } catch (IOException io) {
                    io.printStackTrace();
                } finally {
                    if (outputPipe != null) {
                        outputPipe.close();
                    }
                }
            }


            // Need to look remotely.
            discoverySvc.getRemoteAdvertisements(null, DiscoveryService.ADV, 
                                                 "Name", 
                                                 Teacher.DISTANCE_LEARNING + 
                                                 clientName, 10, null);

            try {
                Thread.sleep(5 * 1000); // 5 sec too short if this is truly distant
                System.out.print(".");
            } catch (Exception ignore) {
            }
        }
    }

    /**
     * Publish our pipe advertisement.
     */
    public void publishPipeAdv() throws IOException {
        PipeAdvertisement pipeAdv = null;
        String advType = PipeAdvertisement.getAdvertisementType();

        pipeAdv = (PipeAdvertisement) AdvertisementFactory.newAdvertisement(
                          advType);

        pipeAdv.setPipeID(IDFactory.newPipeID(netPeerGroup.getPeerGroupID()));
        pipeAdv.setName(DISTANCE_LEARNING + TEACHER_PIPE);
        pipeAdv.setType(PipeService.UnicastType);

        discoverySvc.publish(pipeAdv, DiscoveryService.ADV);
        discoverySvc.remotePublish(pipeAdv, DiscoveryService.ADV);
        this.pipeAdv = pipeAdv;
    }

    /**
     * Create an input pipe for inbound messages.
     */
    public void createInputPipe() throws IOException {
        PipeService pipeSvc = netPeerGroup.getPipeService();
        PipeMsgListener listener = new InputPipeMsgListener();
        this.inputPipe = pipeSvc.createInputPipe(this.pipeAdv, listener);
    }

    /**
     * Start up the Net Peer group & get the Discovery Service for that group.
     */
    public void startJxta() throws PeerGroupException, IOException {
        this.netPeerGroup = PeerGroupFactory.newNetPeerGroup();
        this.discoverySvc = netPeerGroup.getDiscoveryService();


        // Flush all local cache information.  We want to start from scratch.
        discoverySvc.flushAdvertisements(null, DiscoveryService.ADV);
        discoverySvc.flushAdvertisements(null, DiscoveryService.PEER);
        discoverySvc.flushAdvertisements(null, DiscoveryService.GROUP);
    }

    /**
     * Create the Teacher and initiate the JXTA services.
     */
    public static void main(String[] argc) throws Exception {
        Teacher teacher = new Teacher();
        teacher.startJxta();
        teacher.publishPipeAdv();
        teacher.createInputPipe();
        System.out.println("Ready to teach");
    }

    /**
     * Handles the Student Teacher interactions.
     */
    class InputPipeMsgListener implements PipeMsgListener {
        /**
         * Utility method to extract the value of a provided key
         *
         * @param message The Message to parse.
         * @param key The key to search for.
         */
        private String getElementValue(Message message, String key) {
            if (!message.hasElement(key)) {
                return null;
            }

            MessageElement element = message.getElement(key);
            InputStream stream = element.getStream();

            try {
                byte[] bytes = new byte[stream.available()];
                stream.read(bytes);

                return new String(bytes);
            } catch (IOException io) {
                io.printStackTrace();

                return null;
            }
        }

        /**
         * Get the next question from the Agent and send it to the student.
         */
        private void sendNextQuestion(String clientName) {
            Agent agent = (Agent) Teacher.this.agentTable.get(clientName);
            Agent.Problem problem = agent.getNextProblem();
            Map map = new HashMap();

            map.put(Teacher.ACTION_TAG, "QUESTION");
            map.put("question", problem.getQuestion());
            map.put("number", "" + problem.getNumber());
            map.put("total", "" + agent.getTotalNumberOfQuestions());

            try {
                Teacher.this.sendMessage(clientName, map);
            } catch (IOException io) {
                System.out.println("Unable to send a message.  Exiting.");
                io.printStackTrace();
                System.exit(1);
            }
        }

        /**
         * Processes incoming messages from Student objects.
         */
        public void pipeMsgEvent(PipeMsgEvent pipeMsgEvent) {
            Message message = pipeMsgEvent.getMessage();
            String messageAction = getElementValue(message, Teacher.ACTION_TAG);
            String clientName = getElementValue(message, Teacher.CLIENT_NAME);
            String courseName = getElementValue(message, Teacher.COURSE_NAME);

            System.out.println("Received a message from " + clientName);
            System.out.println("tag=" + messageAction);

            if (messageAction.equals("NEW")) {
                // Student requests a new course.
                Agent agent = agentFactory.getAgent(courseName);


                // keep track of the student and agent pair with HashTable
                agentTable.put(clientName, agent);
                sendNextQuestion(clientName);
            } else if (messageAction.equals("ANSWER")) {
                // Process the Student's answer
                Agent agent = (Agent) Teacher.this.agentTable.get(clientName);
                String answer = getElementValue(message, "ANSWER");
                String nextQuestion = (!agent.processAnswer(answer))
                                      ? ("Incorrect.  The correct answer is: " + 
                                        agent.getLastAnswer() + 
                                        ".  You answered: " + answer + "\n")
                                      : "You answered the question correctly.\n";

                Map map = new HashMap();

                if (agent.hasMoreProblems()) {
                    // Send the next question
                    Agent.Problem problem = agent.getNextProblem();
                    nextQuestion += ("The next question is: " + problem.getQuestion());

                    map.put(Teacher.ACTION_TAG, "QUESTION");
                    map.put("question", nextQuestion);
                    map.put("number", "" + problem.getNumber());
                    map.put("total", "" + agent.getTotalNumberOfQuestions());
                } else {
                    // no more problems to solve :)
                    map.put("question", nextQuestion);
                    map.put(Teacher.ACTION_TAG, "END");

                    String suggestion = "You answered " + 
                                        agent.getTotalAnswersCorrect() + 
                                        " out of " + 
                                        agent.getTotalNumberOfQuestions() + 
                                        " questions correctly.\n" + 
                                        agent.getSuggestions();
                    map.put("SUGGESTIONS", suggestion);


                    // Remove the agent from the HashMap
                    Teacher.this.agentTable.remove(agent);
                }

                try {
                    // Send the response back to the Student.
                    Teacher.this.sendMessage(clientName, map);
                } catch (IOException io) {
                    System.out.println("Unable to send a message.  Exiting.");
                    io.printStackTrace();
                    System.exit(1);
                }
            }
        }
    } ///:~ InputPipeMsgListener
} ///:~ Teacher
